'use strict';

/**
 * Various docs for reference
 * https://developer.omniture.com/en_US/content_page/sitecatalyst-tagging/c-tagging-overview
 * http://blogs.adobe.com/digitalmarketing/analytics/custom-link-tracking-capturing-user-actions/
 * https://developer.omniture.com/en_US/forum/general-topic-forum/difference-between-s-t-and-s-t1
  * https://marketing.adobe.com/resources/help/en_US/sc/implement/appmeasure_mjs_migrate.html
 * https://webanalyticsfordevelopers.com/2013/12/10/migrating-from-h-code-to-appmeasurement/
 * https://blogs.adobe.com/digitalmarketing/mobile/lighten-up-with-the-smaller-faster-analytics-javascript-for-web-mobile/
 * https://webanalyticsfordevelopers.com/2013/04/05/the-s_code-js-file-plugins/
 * https://webanalyticsfordevelopers.com/2013/04/09/variables/
 * https://webanalyticsfordevelopers.com/2013/07/02/tracking-links-actions/
 * https://docs.adobe.com/docs/en/aem/6-2/administer/integration/marketing-cloud/sitecatalyst.html
 * https://blogs.adobe.com/digitalmarketing/analytics/custom-link-tracking-capturing-user-actions/
 * https://www.cheatography.com/dmpg-tom/cheat-sheets/adobe-analytics-omniture-sitecatalyst-parameters/
 * http://www.adobe.com/marketing-cloud/web-analytics.html
 * http://www.adobe.com/marketing-cloud/web-analytics/standard.html
 * https://webanalyticsfordevelopers.com/2013/07/16/context-data-variables-processing-rules/
 * https://marketing.adobe.com/developer/documentation/data-insertion/c-visitor-id
 * https://marketing.adobe.com/developer/documentation/data-insertion/c-delayed-data
 * https://marketing.adobe.com/resources/help/en_US/sc/implement/timestamps-overview.html
 * https://marketing.adobe.com/developer/get-started/concepts-terminology/c-what-are-the-available-analytics-variables-and-events
 * https://blogs.adobe.com/digitalmarketing/analytics/products-variable-inside-omniture-sitecatalyst/
 * http://webanalyticssimplified.blogspot.com/2010/02/omniture-sitecatalyst-events.html
 */

/**
 * Dependencies
 */

var integration = require('@segment/analytics.js-integration');
var dot = require('obj-case');
var each = require('@ndhoule/each');
var iso = require('@segment/to-iso-string');
var Track = require('segmentio-facade').Track;

/**
 * hasOwnProperty reference.
 */

var has = Object.prototype.hasOwnProperty;

/**
 * Keys that get applied to `window.s` object,
 * but need to be cleared out before every call.
 *
 * We need to keep this array to share state because otherwise,
 * we can't delete existing keys on the window.s object in between
 * `.track()` and `.page()` calls since they may set different fields.
 * This prevents properties of `.page()` call being sent along with
 * `.track()` calls.
 */

var dynamicKeys = [];

/**
 * Expose `Adobe Analytics`.
 */

var AdobeAnalytics = module.exports = integration('Adobe Analytics');

/**
 * Uses the default `field` from the `window.s` object if it exists,
 * otherwise uses the passed in `value`. Also adds the field to
 * `this.sOptions` and `this.options`
 *
 * @api private
 * @param {string} field
 * @param {*} value
 */

AdobeAnalytics.sOption = function(field, value) {
  var s = window.s;
  var isValid = s && has.call(s, field) && !isEmptyString(field);

  value = isValid ? s[field] : value;

  // TODO: Consider removing this. Not sure why we are doing this since it has no future reference
  this.prototype.sOptions = this.prototype.sOptions || {};
  this.prototype.sOptions[field] = value;

  // Set field and value to this.options
  return this.option(field, value);
};

/**
 * Add our Adobe Analytics instance
 */

AdobeAnalytics
  .global('s')
  .global('s_gi')
  .option('events', {})
  .option('eVars', {})
  .option('props', {})
  .option('hVars', {})
  .option('reportSuiteId', window.s_account)
  .option('timestampOption', 'enabled')
  .option('productIdentifier', 'name')
  .option('marketingCloudOrgId', null)
  .sOption('visitorID')
  .sOption('channel')
  .sOption('campaign')
  .sOption('state')
  .sOption('zip')
  .sOption('pageName')
  .sOption('trackingServer')
  .sOption('trackingServerSecure')
  .sOption('visitorMigrationKey')
  .sOption('visitorMigrationServer')
  .sOption('visitorNamespace')
  .sOption('dc')
  .sOption('charSet', 'ISO-8859-1')
  .sOption('currencyCode', 'USD')
  .sOption('trackDownloadLinks', true)
  .sOption('trackExternalLinks', true)
  .sOption('trackInlineStats', true)
  .sOption('linkDownloadFileTypes', 'exe,zip,wav,mp3,mov,mpg,avi,wmv,pdf,doc,docx,xls,xlsx,ppt,pptx')
  .sOption('linkInternalFilters')
  .sOption('linkLeaveQueryString', false)
  .sOption('linkTrackVars', 'None')
  .sOption('linkTrackEvents', 'None')
  .sOption('usePlugins', true)
  .tag('<script src="//cdn.segment.io/integrations/omniture/AppMeasurement-1.6/appmeasurement.min.js">');

/**
 * Initialize.
 *
 * @api public
 */

AdobeAnalytics.prototype.initialize = function() {
  var options = this.options;
  var self = this;

  // Lowercase all keys of event map for easy matching later
  lowercaseKeys(options.events);

  // In case this has been defined already
  window.s_account = window.s_account || options.reportSuiteId;

  this.load(function() {
    var s = window.s;
    s.trackingServer = s.trackingServer || options.trackingServerUrl;
    s.trackingServerSecure = s.trackingServerSecure || options.trackingServerSecureUrl;

    // visitorAPI.js is included in our rendering of appmeasurement.js to prevent race conditions
    // current .load() function does not guarantee synchronous loading which AA requires
    // visitorAPI is loaded before appmeasurement.js and initialized after both scripts were loaded synchronously
    if (options.marketingCloudOrgId && window.Visitor && typeof window.Visitor.getInstance === 'function') {
      s.visitor = window.Visitor.getInstance(options.marketingCloudOrgId, {
        trackingServer: window.s.trackingServer || options.trackingServerUrl,
        trackingServerSecure: window.s.trackingServerSecure || options.trackingServerSecureUrl
      });
    }
    self.ready();
  });
};

/**
 * Adobe Analytics is loaded if the `window.s_gi` function exists.
 *
 * @return {Boolean} loaded
 */

AdobeAnalytics.prototype.loaded = function() {
  return !!window.s_gi;
};

/**
 * Page.
 *
 * @api public
 * @param {Page} page
 */

AdobeAnalytics.prototype.page = function(page) {
  // Delete any existing keys on window.s from previous call
  clearKeys(dynamicKeys);

  // Set the page name
  var pageName = page.fullName();
  // TODO: for nameless analytics.page(), pageName is `undefined`
  // Should we be setting or sending something else here?
  // When window.s.pageName is not set, AA falls back on url which is bad.
  // Not sure what happens when it is sent as `undefined` (not string)
  // Either way, any change here would be breaking
  window.s.pageName = pageName;

  // Visitor ID aka AA's concept of `userId`.
  // This is not using `update()` so once it is set, it will be sent
  // with `.track()` calls as well.
  // visitorId is not supported for timestamped hits
  // https://marketing.adobe.com/resources/help/en_US/sc/implement/timestamps-overview.html
  var userId = this.analytics.user().id();
  if (userId && this.options.timestampOption === 'disabled') window.s.visitorID = userId;

  // Attach some variables on the `window.s` to be sent with the call
  update(pageName, 'events');
  updateCommonVariables(page, this.options);

  if (this.options.timestampOption !== 'disabled') updateTimestamp(page);

  // Check if any properties match mapped eVar, prop, or hVar in options
  var props = extractProperties(page.properties(), this.options);
  // Attach them to window.s and push to dynamicKeys
  each(update, props);

  // actually make the "page" request, just a single "t" not "tl"
  // "t" will send all variables on the window.s while "tl" does not
  // "t" will increment pageviews while "tl" does not
  window.s.t();
};

/**
 * Track.
 *
 * @api public
 * @param {Track} track
 */

AdobeAnalytics.prototype.track = function(track) {
  // Delete any existing keys on window.s from previous call
  clearKeys(dynamicKeys);

  // Find AA event name from setting's event map
  // otherwise abort
  var adobeEvent = aliasEvent(track.event(), this.options.events);
  if (!adobeEvent) return;

  var properties = track.properties();
  // Tell Adobe Analytics name of the event
  window.s.linkTrackEvents = adobeEvent;
  // Set eVar for the event name if mapped in settings
  var eVarEvent = dot(this.options.eVars, track.event());
  update(track.event(), eVarEvent);

  // Attach some semantic variables on the `window.s` to be sent with the call
  update(adobeEvent, 'events');
  updateCommonVariables(track, this.options);

  // Reporting suites can either be timestamp enabled, timestamp optional or timestamp disabled
  if (this.options.timestampOption !== 'disabled') updateTimestamp(track);

  // Check if any properties match mapped eVar, prop, or hVar in options
  var props = extractProperties(properties, this.options);
  // Attach them to window.s and push to dynamicKeys
  each(update, props);

  // Tell AA which variables (props/eVars/hVars) to track
  // This is only required for `s.tl()` not `s.t()`
  window.s.linkTrackVars = dynamicKeys.join(',');
  // Send request off to Adobe Analytics
  // 1st param: sets 500ms delay to give browser time, also means you are tracking something other than a href link
  // 2nd param: 'o' means 'Other' as opposed to 'd' for 'Downloads' and 'e' for Exit links
  // 3rd param: link name you will see in reports
  window.s.tl(true, 'o', track.event());
};

/**
 * Viewed Product.
 *
 * https://segment.com/docs/spec/ecommerce/#viewed-product
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.productViewed = function(track) {
  processEcommerceEvent(track, 'prodView', this.options);
};


/**
 * Product Added
 *
 * https://segment.com/docs/spec/ecommerce/#added-removed-product
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.productAdded = function(track) {
  processEcommerceEvent(track, 'scAdd', this.options);
};

/**
 * Product Removed
 *
 * https://segment.com/docs/spec/ecommerce/#added-removed-product
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.productRemoved = function(track) {
  processEcommerceEvent(track, 'scRemove', this.options);
};

/**
 * Order Completed
 * https://segment.com/docs/spec/ecommerce/#completing-an-order
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.orderCompleted = function(track) {
  processEcommerceEvent(track, 'purchase', this.options);
};

/**
 * Cart Viewed
 *
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.cartViewed = function(track) {
  processEcommerceEvent(track, 'scView', this.options);
};

/**
 * Checkout Started
 *
 * @api public
 * @param {Track} Track
 */

AdobeAnalytics.prototype.checkoutStarted = function(track) {
  processEcommerceEvent(track, 'scCheckout', this.options);
};

/**
 * Process Ecommerce events
 *
 * @api private
 * @param {Facade} track
 * @param {String} adobe event name
 * @param {Object} options
 */

function processEcommerceEvent(track, name, options) {
  var props = track.properties();
  var identifier = options.productIdentifier; // can be 'name', 'sku', or 'id' with first being default
  var productVariables = '';

  // Iterate over products for order completed & cart events
  if (name === 'purchase' || name === 'scView') {
    var products = track.products();
    var productDescription;

    // Adobe Analytics wants product description in semi-colon delimited string separated by commas
    for (var x = 0; x < products.length; x++) {
      var product = new Track({ properties: products[x] });
      productDescription = formatProduct(product, identifier);
      productVariables += productDescription;
      // if there are more products, delimit using comma
      if (products[x + 1]) productVariables += ',';
    }
  } else if (name === 'scRemove' || name === 'scAdd' || name === 'prodView') {
    // for these events, there is only one product
    productVariables = formatProduct(track, identifier);
  }
  // NOTE: checkout event (scCheckout) don't have product data

  var adobeVariables = extractProperties(props, options);
  var eVarEvent = dot(options.eVars, track.event());

  // Removing any existing properties on window.s from previous calls
  // and attach new properties to send off
  clearKeys(dynamicKeys);
  updateCommonVariables(track, options);
  update(name, 'events');
  update(productVariables, 'products');
  update(track.event(), eVarEvent);
  each(update, adobeVariables);

  // for orderCompleted events only
  // Allows AA to track offline data like refunds, cancelled order
  if (name === 'purchase') {
    update(props.purchaseId || track.orderId(), 'purchaseID');
    update(props.transactionId || track.orderId(), 'transactionID');
  }

  if (options.timestampOption !== 'disabled') updateTimestamp(track);

  window.s.linkTrackEvents = name;
  window.s.linkTrackVars = dynamicKeys.join(',');

  // Actually fire the event off to AA!
  window.s.tl(true, 'o', track.event());
}

/**
 * Update a number of Adobe Analytics common variables on window.s to be
 * sent with the call
 *
 * @api private
 * @param {Facade} facade
 */

function updateCommonVariables(facade, options) {
  var properties = facade.properties();

  update(properties.channel || options.channel, 'channel');
  update(properties.campaign || options.campaign, 'campaign');
  update(properties.state || options.state, 'state');
  update(properties.zip || options.zip, 'zip');
  if (facade.type() === 'track') update(properties.pageName || options.pageName || facade.proxy('context.page.title'), 'pageName');
}

/**
 * Update the AA variable and explicitly mark the link vars
 *
 * @api private
 * @param {String, String} value, key
 * @return {string} window.s[key]
 */

function update(value, key) {
  // Make sure we have everything we need
  if (!key || value === undefined || value === null || value === '') return;
  // Push each key to be sent later with the payload
  // Only keys listed here will be parsed from the window.s object when data is sent via s.tl() aka `.track()` calls
  dynamicKeys.push(key);
  // Set Adobe Analytics variables on the window.s to be sent
  window.s[key] = value.toString();
}

/**
 * Update window.s.timestamp
 *
 * @api private
 * @param {Facade} facade
 */

function updateTimestamp(facade) {
  var properties = facade.properties();
  var timestamp = properties.timestamp || facade.timestamp();
  // Timestamp must be either unix or iso-8601
  if (typeof timestamp !== 'string') timestamp = iso(timestamp);
  update(timestamp, 'timestamp');
}

/**
 * Alias a regular event `name` to an AA event, using a dictionary of
 * `events`.
 *
 * @api private
 * @param {string} name
 * @param {Object} events
 * @return {string|null}
 */

function aliasEvent(name, events) {
  if (events[name.toLowerCase()]) return events[name.toLowerCase()];
}

/**
 * Format semantic ecommerce product properties to Adobe Analytics variable strings.
 *
 * @api private
 * @param {Object} props
 * @return {string}
 */

function formatProduct(props, identifier) {
  var quantity = props.quantity() || 1;
  var total = ((props.price() || 0) * quantity).toFixed(2);
  return [
    props.category(),
    props[identifier](),
    quantity,
    total
  ].join(';');
}

/**
 * Clear last keys used with Adobe Analytics.
 *
 * @api private
 * @param {Array} keys
 */

function clearKeys(keys) {
  each(function(linkVar) {
    delete window.s[linkVar];
  }, keys);
  // Clears the array passed in
  keys.length = 0;
}

/**
 * Extract properties for `window.s`.
 *
 * @api private
 * @param {Object} props
 * @param {Object} options
 */

function extractProperties(props, options) {
  var result = {};
  var mappings = [options.eVars, options.props, options.hVars];

  // Iterate through each variable mappings to find matching props
  for (var x = 0; x < mappings.length; x++) {
    each(match, mappings[x]);
  }
  function match(mappedValue, mappedKey) {
    var value = dot(props, mappedKey);
    // make sure it's an acceptable data type
    if (value !== null && !isFunction(value) && typeof value !== 'object') result[mappedValue] = value;
  }

  return result;
}

/**
 * Check if function is a function
 *
 * @api private
 * @param {Object} props
 * @param {Object} options
 */

function isFunction(fn) {
  var getType = {};
  return fn && getType.toString.call(fn) === '[object Function]';
}

/**
 *
 * @api private
 * @param {Object} obj
 * @return {Object}
 */

function lowercaseKeys(obj) {
  obj = obj || {};
  each(function(value, key) {
    delete obj[key];
    obj[key.toLowerCase()] = value;
  }, obj);
  return obj;
}

/**
 * Return whether `str` is an empty string.
 *
 * @api private
 * @return {boolean} empty
 */

function isEmptyString(str) {
  return typeof str === 'string' && str.length === 0;
}
